/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_TEE_H__
#define __GDAM_TEE_H__

typedef struct _GdamTeeSink GdamTeeSink;
typedef struct _GdamTee GdamTee;
typedef struct _GdamTeeSinkClass GdamTeeSinkClass;

#include <gtk/gtkobject.h>
#include "gdamsource.h"
#include <gdam/gdambuffer.h>
#include <gdam/gdamsample.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


GtkType gdam_tee_sink_get_type();
GtkType gdam_tee_get_type();
#define GDAM_TYPE_TEE_SINK			(gdam_tee_sink_get_type ())
#define GDAM_TEE_SINK(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_TEE_SINK, GdamTeeSink))
#define GDAM_TEE_SINK_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAM_TYPE_TEE_SINK, GdamTeeSinkClass))
#define GDAM_IS_TEE_SINK(obj)           (GTK_CHECK_TYPE ((obj), GDAM_TYPE_TEE_SINK))
#define GDAM_IS_TEE_SINK_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAM_TYPE_TEE_SINK))
#define GDAM_TYPE_TEE              (gdam_tee_get_type ())
#define GDAM_TEE(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_TEE, GdamTee))
#define GDAM_TEE_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAM_TYPE_TEE, GdamTeeClass))
#define GDAM_IS_TEE(obj)           (GTK_CHECK_TYPE ((obj), GDAM_TYPE_TEE))
#define GDAM_IS_TEE_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAM_TYPE_TEE))


struct _GdamTeeSinkClass {
	GdamSinkClass		sink_class;
};

struct _GdamTeeClass {
	GdamSourceClass		source_class;
};

/*
 * Tee Sink - the end of the tee that holds a single "feed" source
 * and a buffer 
 */
struct _GdamTeeSink {
	/* sink end of the `tee'. */
	GdamSink  	    sink;

	/* number of samples we've read from feed. */
	GdamTime            samples_read;

	int		    id;

	/* The buffer that all the GdamTees read out of. */
	GdamSample*	    buffer;

	/* The next index that the feed should write to. */
	int                 read_index;

	/* The number of bytes in the buffer. */
	int                 num_buffered;

	/* The allocated size of the buffer (also the max amount
	 * the sources can use). */
	int                 buffer_size;
};

/*
 * Tee - a source that gets its data from a TeeSink.
 */
struct _GdamTee {
	GdamSource          source;
	GdamTeeSink*        tee_sink;
	/* Time the next sample output by this sink should
	 * be at. */
	GdamTime            tee_time;
};


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
