/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_EQUALIZER_H_
#define __GDAM_EQUALIZER_H_

#include <gdam/gdamfilter.h>

GtkType gdam_equalizer_get_type();
#define GDAM_TYPE_EQUALIZER   gdam_equalizer_get_type()
#define GDAM_EQUALIZER(ptr) ((GdamEqualizer*) gtk_type_check_object_cast ((GtkTypeObject*) (ptr), GDAM_TYPE_EQUALIZER))

/*
 * Implements a simple audio filter,
 * as described in "The Audio-EQ Cookbook",
 * by Robert Bristow-Johnson, pbjrbj@viconet.com.
 * which is widely circulated on the web.
 */

typedef enum _GdamEqualizerMode GdamEqualizerMode;
typedef struct _GdamEqualizer GdamEqualizer;

enum _GdamEqualizerMode {
	GDAM_EQUALIZER_LOW_PASS = 300,
	GDAM_EQUALIZER_HIGH_PASS,
	GDAM_EQUALIZER_BAND_PASS,
	GDAM_EQUALIZER_NOTCH,
	GDAM_EQUALIZER_PEAKING_EQ,
	GDAM_EQUALIZER_LOW_SHELF,
	GDAM_EQUALIZER_HIGH_SHELF,
};


struct _GdamEqualizer {
	GdamFilter			filter;

	/* y[n] = x_coeff[0] * x[n] 
	        + x_coeff[1] * x[n-1] 
	        + x_coeff[2] * x[n-2] 
	        + y_coeff[1] * y[n-1] 
	        + y_coeff[2] * y[n-2]           [where x is input, y is output]
	 */
	GdamFixed			x_coeff[3];
	GdamFixed			y_coeff[3];
	GdamSample			x[2];
	GdamSample			y[2];

	int				effect;

	gdouble                         frequency;
	gdouble				bandwidth;
	gdouble				db_gain;

	GdamEqualizerMode		type;
};

#endif
