/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_BUFFER_VIEWPORT_H_
#define __GDAM_BUFFER_VIEWPORT_H_


typedef struct _GdamBufferViewport GdamBufferViewport;
typedef struct _GdamBufferViewportClass GdamBufferViewportClass;
typedef struct _GdamBufferViewportMarker GdamBufferViewportMarker;

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include "gdamskin.h"

GtkType gdam_buffer_viewport_get_type();
#define GDAM_TYPE_BUFFER_VIEWPORT			(gdam_buffer_viewport_get_type ())
#define GDAM_BUFFER_VIEWPORT(obj)              (GTK_CHECK_CAST ((obj), GDAM_TYPE_BUFFER_VIEWPORT, GdamBufferViewport))
#define GDAM_BUFFER_VIEWPORT_CLASS(klass)      (GTK_CHECK_CLASS_CAST ((klass), GDAM_TYPE_BUFFER_VIEWPORT, GdamBufferViewportClass))
#define GDAM_IS_BUFFER_VIEWPORT(obj)           (GTK_CHECK_TYPE ((obj), GDAM_TYPE_BUFFER_VIEWPORT))
#define GDAM_IS_BUFFER_VIEWPORT_CLASS(klass)   (GTK_CHECK_CLASS_TYPE ((klass), GDAM_TYPE_BUFFER_VIEWPORT))

struct _GdamBufferViewportMarker {
	int			draw_x, draw_y;
	int			draw_width;
	char*			text;
};
struct _GdamBufferViewportClass {
	GdamSkinClass	skin_class;
	void          (*selection_changed)(GdamBufferViewport* viewport);
	void          (*view_changed)     (GdamBufferViewport* viewport);
	void          (*beat_lines_changed)(GdamBufferViewport* viewport);
	void          (*length_changed)   (GdamBufferViewport* viewport);
};
struct _GdamBufferViewport {
	GdamSkin		skin;

	/* Buffer that we are viewing, if any. */
	guint			buffer_id;

	/* Height and width of the buffer in pixels. */
	int			draw_width, draw_height;

	/* Drawing area is `main_widget': it must be a GtkEventBox. */
	GdkGC*			fg_gc;
	GdkGC*			bg_gc;

	/* Marker drawing information. */
	GdkFont*		marker_font;
	GdkGC*			marker_gc;
	GdkColor*               marker_color;
	char*                   marker_format;
	gdouble			marker_unit;
	int			num_markers;
	GdamBufferViewportMarker* markers;

	/* Usually 44100 but the beat_calculator uses this... */
	int			samples_per_second;

	/* Beat lines. */
	gboolean		can_drag_beat_lines;
	gboolean		is_dragging_beat_line;
	int			dragging_beat_line;
	gdouble			start_beat;
	gdouble			beat_period;
	int			num_beats;
	GdkGC*			beat_line_gc;
	
	/* Are we using the buffer manager for this buffer? */
	gboolean		use_buffer_manager;

	/* Sample length, if not using the buffer manager. */
	gdouble			sample_length;

	/* Start/end times for the view. */
	gdouble			time_start;
	gdouble			time_length;

	/* Where a selection is, if it exists. */
	gboolean		has_selection;
	gdouble			selection_start;
	gdouble			selection_length;

	/* Whether a query of the server is outstanding. */
	gboolean		getting_profile;

	/* If a query is outstanding, don't query again if the size
	 * changes -- instead set this flag to query the server
	 * once the next query comes in.  This prevents backlogging.  */
	gboolean		redo_query;

	/* If nonzero, a redraw is pending. */
	guint			redraw_thread;

	gboolean		is_dragging;
	gboolean		is_selectable;
	gdouble			last_drag;
	gdouble			drag_start;


	/* Number of dots to capture per output pixel. */
	int			density;
	int			num_last_samples;
	gint*			min_last_samples;
	gint*			max_last_samples;

};

/* Create a new GdamBufferViewport -- you *must* use this function
 * to leave the viewport in a usable state. */
GdamBufferViewport* 
           gdam_buffer_viewport_new_shown     (GdamChannel*	   channel);

void       gdam_buffer_viewport_set_buffer_id (GdamBufferViewport* viewport,
                                               guint               buffer_id);
void       gdam_buffer_viewport_set_view      (GdamBufferViewport* viewport,
                                               gdouble             start,
					       gdouble             length);
void       gdam_buffer_viewport_set_selection (GdamBufferViewport* viewport,
                                               gdouble             start,
					       gdouble             length);
void       gdam_buffer_viewport_set_zoom      (GdamBufferViewport* viewport,
                                               gdouble             zoom);
void       gdam_buffer_viewport_set_center    (GdamBufferViewport* viewport,
                                               gdouble             zoom);
void       gdam_buffer_viewport_maybe_uncache(GdamBufferViewport* viewport);

/* Maybe cause requeries of server -- possibly blocked by backlog
 * protection. */
void       gdam_buffer_viewport_buffer_changed(GdamBufferViewport* viewport);
void       gdam_buffer_viewport_changed       (GdamBufferViewport* viewport);

/* Snap the window to the exact length of the sample. */
void       gdam_buffer_viewport_whole_sample  (GdamBufferViewport* viewport);

/* Private: initialize the main_widget to the GtkEventBox.
 * (It doesn't get shown...). */
void       gdam_buffer_viewport_new_event_box (GdamBufferViewport* viewport);

void       gdam_buffer_viewport_queue_redraw  (GdamBufferViewport* viewport);

void       gdam_buffer_viewport_set_length    (GdamBufferViewport* viewport,
                                               gdouble             length);
void gdam_buffer_viewport_beat_lines_changed(GdamBufferViewport* viewport);
#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
