/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamturntableskin.h"
#include <gdam/component/gdamproducer.h>
#include <stdlib.h>
#include <string.h>
#include <gdk/gdk.h>
#include <glade/glade.h>
#include <gtk/gtkdnd.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkrange.h>
#include <gtk/gtklabel.h>
#include <gdam/component/gdamutil.h>
#include "gdamdnd.h"
#include "gdamskin.h"
#include "gdamskinutil.h"
#include "gdamkeyboardtrap.h"

static gboolean gdam_turntable_build_new_model(GdamSkin*      skin,
					       GdamChannel*   channel,
					       int            num_options,
				               char**         options);
static void gdam_turntable_skin_synchronize   (GdamView*      view);

/*    ____ _                 ____  _          __  __ 
 *   / ___| | __ _ ___ ___  / ___|| |_ _   _ / _|/ _|
 *  | |   | |/ _` / __/ __| \___ \| __| | | | |_| |_ 
 *  | |___| | (_| \__ \__ \  ___) | |_| |_| |  _|  _|
 *   \____|_|\__,_|___/___/ |____/ \__|\__,_|_| |_|  
 */                                                 

static void gdam_turntable_skin_class_init(GdamTurntableSkinClass* klass);

static void gdam_turntable_skin_init(GdamTurntableSkin* skin)
{
}

GtkType gdam_turntable_skin_get_type()
{
	static GtkType turntable_skin_type = 0;
	if (!turntable_skin_type) {
		static const GtkTypeInfo turntable_skin_info = {
			"GdamTurntableSkin",
			sizeof(GdamTurntableSkin),
			sizeof(GdamTurntableSkinClass),
			(GtkClassInitFunc) gdam_turntable_skin_class_init,
			(GtkObjectInitFunc) gdam_turntable_skin_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		turntable_skin_type = gtk_type_unique(GDAM_TYPE_SKIN,
		                                     &turntable_skin_info);
	}
	return turntable_skin_type;
}


/*  ____  _    _         _____                 _   _                 
 * / ___|| | _(_)_ __   |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
 * \___ \| |/ / | '_ \  | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
 *  ___) |   <| | | | | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
 * |____/|_|\_\_|_| |_| |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
 *
 */                                                                 

#if 0
static void on_position_string_drop(const char* 	new_position_str,
				    gpointer		user_data)
{
	GdamTurntable* tt = GDAM_TURNTABLE(user_data);
	float new_position = atof(new_position_str);
	GDAM_PRODUCER(tt)->stopped_position = new_position;
	turntable_refresh_time_displays(tt, new_position);
		/* TODO: what if playing? */
}
#endif

static void on_tempo_string_drop(const char* 		new_bpm_str,
				 gpointer		user_data)
{
	GdamTurntable* tt = GDAM_TURNTABLE(user_data);
	float new_bpm = atof(new_bpm_str);
	if (new_bpm > 0)
		    turntable_set_tempo(tt, new_bpm);
}

static void on_tempo_string_drag(GtkSelectionData* 	selection_data,
				 gpointer		user_data)
{
	GdamTurntable* tt = GDAM_TURNTABLE(user_data);
	GdamModel* component = GDAM_MODEL(tt);
	char str[256];
	sprintf(str, "%f", gdam_beat_info_get_bpm(&component->beat_info,0));
	gtk_selection_data_set(selection_data,
			       selection_data->target,
			       8,
			       (guchar*)str,
			       strlen(str));
}

/*  This function is called regularly to keep the sync preview up to date
 */ 
static int auto_sync_timeout(GdamTurntableSkin* skin)
{
	if (!GDAM_SKIN(skin)->model->source_id){
		skin->timer_id = 0;
		return FALSE;
	}
	turntable_do_sync_preview(GDAM_TURNTABLE(GDAM_SKIN(skin)->model));
	return TRUE;
}

/*  Callback for 'PLAY' button, plays the song.
 */
static void on_play_button_clicked(GtkButton*		button, 
			           GdamTurntableSkin* 	skin)
{
	GdamTurntable*  turntable = GDAM_TURNTABLE(GDAM_SKIN(skin)->model);
	gdouble position;
	GdamPlayInfo* pi;
	if (((GdamModel*)turntable)->source_id != 0) {
		turntable_do_sync(turntable);
	 	return;
	}
	pi = g_new0(GdamPlayInfo, 1);
	/* Get the turntable's position. (position is a member of producer). */
	position = GDAM_PRODUCER(turntable)->stopped_position;
	if ((skin->event_state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK) {
		guint leader_id;
		GdamBeatInfo beat_info;
		gdam_debug("turntable skin: attempting to start in sync");
		if (!gdam_infomanager_get_sync_info(GDAM_MODEL(turntable),
						    &leader_id,
						    &beat_info)) {
			g_warning("couldn't get_sync_info in "
				"play_button_clicked");
			return;
		}
		pi->lead_id = leader_id;
		pi->lead_beat_info = &beat_info;
		pi->position = position;
		pi->is_synchronized = TRUE;
		/* TODO: finish starting in sync */
	} else {
		gdam_debug("turntable skin: not attempting to start in sync");
		pi->position = position;
	}
	gdam_producer_play(GDAM_PRODUCER(turntable), pi);
	skin->timer_id = gtk_timeout_add(500, (GtkFunction)auto_sync_timeout,
					      skin);
}

/*  Callback for button events which trigger a nudge.
 */
static void on_nudge_sec_clicked(GtkObject* object, 
				 NudgeInfo* nudge_info)
{
	if (!nudge_info->turntable)
		return;
	gdam_turntable_seek(nudge_info->turntable, nudge_info->amount);
}

/*  Callback for button events which trigger a measured jump.
 */
static void on_nudge_bar_clicked(GtkObject*      object,
				 NudgeInfo*      nudge_info)
{
	float amount = nudge_info->amount;
	gdam_debug("tt: nudge_by_bar: %f", amount);
	if (!nudge_info->turntable)
		return;

	if (GDAM_IS_KEYBOARD_TRAP(object)){
		gdam_turntable_seek_by_beats(nudge_info->turntable, 4*amount);
		return;
	}

	if ((nudge_info->skin->event_state & GDK_CONTROL_MASK) 
						!= GDK_CONTROL_MASK){
		amount *= 4.0;
	}

	gdam_turntable_seek_by_beats(nudge_info->turntable, amount);
}

/*  Callback for button events which trigger a measured jump.
 */
static void on_nudge_beat_clicked(GtkObject*      object,
				  NudgeInfo*      nudge_info)
{
	float amount = nudge_info->amount;
	gdam_debug("tt: nudge_by_beat: %f", amount);
	if (!nudge_info->turntable)
		return;
	gdam_turntable_seek_by_beats(nudge_info->turntable, amount);
}

static void on_zero_button_clicked(GtkButton*      button, 
	                          GdamTurntable*  turntable)
{
	gdam_turntable_zero(turntable);
}

static void on_start_button_clicked(GtkButton*      button, 
	                           GdamTurntable*  turntable)
{
	if (turntable->producer) {
		const GdamBeatInfo* info;
		gdouble first_beat;
		GdamPlayInfo play_info = {0};
		info = & (GDAM_MODEL(turntable->producer)->beat_info);
		first_beat = gdam_beat_info_get_first_beat(info,0);
		play_info.position = (GdamTime)first_beat;
		gdam_producer_play(turntable->producer, 
				   &play_info);
	}
}

static void on_stop_button_clicked(GtkButton*      button, 
                                   GdamTurntable*  turntable)
{
	if (GDAM_MODEL(turntable)->source_id != 0)
		gdam_producer_stop(turntable->producer);
	else if (turntable->producer) {
		gdouble first_beat;
		const GdamBeatInfo* info;
		info = & (GDAM_MODEL(turntable->producer)->beat_info);
		first_beat = gdam_beat_info_get_first_beat(info,0);
		gdam_producer_set_position(turntable->producer, 
					   first_beat);
	}
}

static void resample_adj_changed(GtkAdjustment* adj, GdamTurntable* turntable)
{
	gdam_debug("turntableskin: resample_adj_changed");
	gdam_turntable_set_rate(turntable, adj->value);
}

static void resample_adj_button_event(GtkWidget*     wid, 
				      GdkEvent*      ev, 
				      GdamTurntableSkin* skin)
{
	if (ev->type == GDK_BUTTON_PRESS) {
		skin->user_dragging_resample = 1;
	} else if (ev->type == GDK_BUTTON_RELEASE) {
		skin->user_dragging_resample = 0;
	}
}

static void on_jump_adj_changed(GtkAdjustment* adj, GdamTurntable* turntable)
{
	gdam_debug("tt:jump_adj changed - %f", adj->value);
	gdam_turntable_update_jump_preview(turntable, 
					   (double)abs((int)(adj->value)));
}

static void
on_jump_button_event(GtkWidget* widget, GdkEvent* ev, GdamTurntable* turntable)
{
	GtkRange* range = GTK_RANGE(widget);
	GtkAdjustment* adj = GTK_ADJUSTMENT(gtk_range_get_adjustment(range));
	if (ev->type == GDK_BUTTON_RELEASE && adj) {
		gfloat jump = adj->value;
		gdam_turntable_seek_by_beats(turntable, (int)jump);
		adj->value = 0;
		gtk_adjustment_changed(adj);
		gdam_turntable_update_jump_preview(turntable, 0);
	}
}

static void position_adj_changed(GtkAdjustment* adj, GdamTurntable* turntable)
{
	GdamModel* model = GDAM_MODEL(turntable);
	if (turntable->user_dragging_position) {
		gdam_producer_set_position(turntable->producer, 
					   adj->value);
		return;
	}
	if (model->source_id && turntable->producer != NULL){
		GdamPlayInfo play_info = {0};
		play_info.position = adj->value;
		gdam_producer_play(turntable->producer, &play_info);
	}
	else {
		gdam_producer_set_position(turntable->producer,
					   adj->value);
	}
}


static void posscroll_button_event(GtkWidget*     wid, 
			    GdkEvent*      ev, 
			    GdamTurntable* turntable)
{
	if (ev->type == GDK_BUTTON_PRESS) {
		turntable->user_dragging_position = 1;
	} else if (ev->type == GDK_BUTTON_RELEASE) {
		if (turntable->producer 
		    && GDAM_MODEL(turntable->producer)->source_id) {
				/* TODO:  this action must be deferred,
				 *        as it is not good here... */
			gdam_producer_seek(GDAM_PRODUCER(turntable), 
	                                   GDAM_PRODUCER(turntable)
					   		->stopped_position, 0);
		}
		turntable->user_dragging_position = 0;
	}
}

/*  These are from infoclient code...
 *
 */
static void
on_anchor_rbutton_toggled(GtkToggleButton*      button,
			  GdamTurntable*  	turntable)
{
	turntable->anchored = gtk_toggle_button_get_active(button);
	gdam_debug("tt:rbutton_toggled  %s", 
				turntable->anchored ? "true" : "false");
	if (turntable && turntable->anchored){
	       gdam_infomanager_set_lead_model(GDAM_MODEL(turntable));
	} 
}

static void on_info_button_clicked(GtkButton*      button,
				   GdamTurntable*  turntable)
{
    g_warning("Ignoring button press.\n");
}

static void
on_tempo_button_clicked(GtkButton* button,
		        GdamTurntable* turntable)
{
	float new_bpm = 0;
	new_bpm = gdam_infomanager_get_desired_bpm((GdamModel*)turntable);
	if (new_bpm > 0.0)
		gdam_infomanager_set_desired_bpm((GdamModel*)turntable,
		                                 new_bpm);
}

static void on_sync_preview_button_clicked(GtkButton*      button,
					   GdamTurntable*  turntab)
{
	turntable_do_sync_preview(turntab);
}

static void on_halve_check_button_clicked(GtkButton* button,
					  GdamTurntable* turntable)
{
	GdamBeatInfo new_beat;
	const GdamBeatInfo* prod_beat;
	float bpm;
	float fb;
	if (GTK_TOGGLE_BUTTON(button)->active)
		turntable->bpm_factor = 2;
	else
		turntable->bpm_factor = 1;
	if (!turntable->producer)
		return;
	prod_beat = &(GDAM_MODEL(turntable->producer)->beat_info);
	bpm = gdam_beat_info_get_bpm(prod_beat,0);
	bpm = bpm * turntable->rate / turntable->bpm_factor;
	fb = gdam_beat_info_get_first_beat(prod_beat,0);
	fb = fb / turntable->rate;
	gdam_beat_info_set_bpm(&new_beat, bpm);
	gdam_beat_info_set_first_beat(&new_beat, fb);
	gdam_model_set_beat_info(GDAM_MODEL(turntable), &new_beat);
}

static void
on_sync_button_clicked(GtkButton*      button,
		       GdamTurntable*  turntable)
{
	turntable_do_sync(turntable);
}


/*
 * Functions to update display widgets
 *
 */
void turntable_update_pos_label(GtkLabel* label, gdouble _new_time,
					GtkObject* object)
{
	char text[63];
	GdamTurntable* turntable = GDAM_TURNTABLE(object);
	gdouble new_time = _new_time * turntable->rate;
	int secs = (int)(new_time);
	int centisecs = (int)((new_time-(float)secs) * 100.0);
	if (new_time < 0)
		sprintf(text, "-%02d:%02d.%02d",
		        secs/60,
			abs(secs%60),
			abs(centisecs));
	else
		sprintf(text, "%03d:%02d.%02d", secs/60, secs%60, centisecs);
	gtk_label_set_text(label, text);
}

static void  turntable_update_beatnum(GtkLabel* label, gdouble new_beatnum,
					GtkObject* object)
{
	char text[63];
	sprintf(text, "%3.3f", new_beatnum);
	gtk_label_set_text(label, text);
}

static void  turntable_update_tempo(GtkLabel* label, const GdamBeatInfo* info,
					GtkObject* object)
{
	char text[63];
	gdam_debug("setting tempo label to %f", gdam_beat_info_get_bpm(info,0));
	sprintf(text, "%3.3f", gdam_beat_info_get_bpm(info,0));
	gtk_label_set_text(label, text);
}

static void  on_turntable_set_rate(GtkLabel* label, gdouble new_rate,
					GtkObject* object)
{
	char text[63];
	gdam_debug("setting rate label to %f", new_rate);
	sprintf(text, "%3.3f%%", new_rate*100);
	gtk_label_set_text(label, text);
}

static void  turntable_update_rate_adj(GdamTurntable* turntable, 
				       gdouble new_rate,
				       GdamTurntableSkin* skin)
{
	if (skin->resample_adj && !skin->user_dragging_resample){
		gdam_debug("setting rate adj to %f", new_rate);
		skin->resample_adj->value = new_rate;
		gtk_adjustment_changed(skin->resample_adj);
	}
	else{
		gdam_debug("declining to set rate to %f", new_rate);
	}
}

static void  update_anchored_label(GtkLabel* label, int anchored,
				   GtkObject* object)
{
	GdamTurntable* tt = GDAM_TURNTABLE(object);
	if (tt && tt->anchored) {
		char text[63];
		sprintf(text, "-");
		gtk_label_set_text(label, text);
	}
}

static void update_sync_preview_label(GtkLabel* label, gdouble new_val,
					GtkObject* object)
{
	GdamTurntable* tt = GDAM_TURNTABLE(object);
	if (tt && !tt->anchored) {
		char text[63];
		sprintf(text, "%1.3f", new_val);
		gtk_label_set_text(label, text);
	}
}

static void  turntable_update_jump_preview(GtkLabel* label, gdouble new_val,
						GtkObject* object)
{
	char text[63];
	gdam_debug("tt:jump_prev - %f", new_val);
	sprintf(text, "%2.0f", new_val);
	gtk_label_set_text(label, text);
}

static void on_tempo_press_event(GtkWidget* widget, GdkEventButton* event,
				 gpointer udata)
{
	if (event->type == GDK_2BUTTON_PRESS)
	{
		GdamModel* model = GDAM_SKIN(udata)->model;
		GdamTurntable* turntable = GDAM_TURNTABLE(model);
		float bpm =  gdam_infomanager_get_desired_bpm(model);
		if (bpm <= 0 || !turntable){
			/*failure*/
			return;
		}
		turntable_set_tempo(turntable, bpm);
	}
}

static void turntable_update_position_adj_value(GtkAdjustment* adj, 
					gdouble new_val, GtkObject* object)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(object);
	if (turntable && !turntable->user_dragging_position && adj){
		adj->value = new_val * turntable->rate;
		if (adj->value > adj->upper) adj->value = adj->upper;
		else if (adj->value < adj->lower) adj->value = adj->lower;
		gtk_adjustment_changed(adj);
	}
}

static void turntable_update_position_adj_upper(GtkAdjustment* adj, 
					gdouble new_val, GtkObject* object)
{
	GdamTurntable* turntable = GDAM_TURNTABLE(object);
	if (turntable && adj){
		adj->upper = new_val * turntable->rate;
		gtk_adjustment_changed(adj);
	}
}

/* A simple callback which traps the state of the most recent button_press callback */
static gboolean any_pressed_callback(GtkObject*         object,
                                     GdkEventButton*    event,
				     GdamTurntableSkin* skin)
{
	skin->event_state = event->state;
	return FALSE;
}

/* Connects the clicked signal of the button to the handler, and also connects 
 * to the button_press_event signal, in order to trap the state of the event 
 * which produced the click. */
static void connect_button_with_mask(GdamSkin*          skin, 
                                     GtkObject*         button,
				     const char*        signal_name,
				     GtkSignalFunc      clicked_callback,
				     gpointer           user_data)
{
	gtk_signal_connect(button,
	                   signal_name,
			   clicked_callback,
			   user_data);
	if (!GTK_IS_WIDGET(button))
		return;
	gtk_signal_connect(button,
	                   "button_press_event",
			   GTK_SIGNAL_FUNC(any_pressed_callback),
			   skin);
}

gboolean gdam_turntable_skin_set_producer(	GdamTurntableSkin* 	skin,
						const char* 	skin_type_name, 
						const char* 	skin_filename,
						int             num_options,
						char**          options)
{
	GdamModel* submodel = NULL;
	GdamSkin* submodel_skin = NULL;
	GdamTurntable* turntable = GDAM_TURNTABLE(GDAM_SKIN(skin)->model);
	if (!turntable || !skin_type_name || !skin_filename) {
		return FALSE;
	}
	/*gdam_debug("gdam_turntable_skin_set_producer: "
	           "type=%s, filename=%s, options=%s",
		     skin_type_name, skin_filename, options);*/

	submodel_skin = gdam_skin_new_by_type_name
	                             (skin_type_name,
                                      skin_filename,
				      num_options,
				      options,
				      GDAM_MODEL(turntable)->channel,
				      &submodel);
	if (!GDAM_IS_PRODUCER(submodel)) {
		g_warning("sorry, submodel of turntable was not "
			  "a producer");
		return FALSE;
	}
	gdam_turntable_set_producer(turntable, GDAM_PRODUCER(submodel));

#if 0
#endif

	if (skin->producer_container) {
		GtkWidget* widget = submodel_skin->main_widget;
		gdam_set_container(skin->producer_container, widget);
	}

	skin->producer_skin = submodel_skin;

	return TRUE;
}

/*  Connect callbacks to objects....  this function is used by gladelib
 *  to connect calback to objects at runtime.  We cannot let libglade 
 *  handle signal connection automatically, because we can't provide
 *  the user_data argument in that case.
 */
static gboolean gdam_turntable_frame_drag_data_received_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gdam_widget_string_dest_set(GTK_WIDGET(object), 
				    on_tempo_string_drop, 
				    "gdam-tempo-text", 
				    GDAM_TURNTABLE(skin->model));
	return TRUE;
}
static gboolean gdam_turntable_nudge_sec_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	float amount;
	char* endp;
	NudgeInfo* nudge_info;
	if (handler_names[1] == NULL) {
		g_warning("%s needs a real-valued argument", handler_names[0]);
		return TRUE;
	}
	amount = strtod(handler_names[1], &endp);
	if (endp == handler_names[1] || *endp != '\0')
		return TRUE;

	nudge_info = g_new(NudgeInfo, 1);
	nudge_info->turntable = GDAM_TURNTABLE(skin->model);
	nudge_info->skin = GDAM_TURNTABLE_SKIN(skin);
	nudge_info->amount = amount;

	connect_button_with_mask(skin, object,
	                         signal_name,
				 GTK_SIGNAL_FUNC(on_nudge_sec_clicked),
				 nudge_info);
	return TRUE;
}

static gboolean gdam_turntable_nudge_bar_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	float amount;
	char* endp;
	NudgeInfo* nudge_info;
	if (handler_names[1] == NULL) {
		g_warning("%s needs a real-valued argument", handler_names[0]);
		return TRUE;
	}
	amount = strtod(handler_names[1], &endp);
	if (endp == handler_names[1] || *endp != '\0')
		return TRUE;

	nudge_info = g_new(NudgeInfo, 1);
	nudge_info->turntable = GDAM_TURNTABLE(skin->model);
	nudge_info->skin = GDAM_TURNTABLE_SKIN(skin);
	nudge_info->amount = amount;
	connect_button_with_mask(skin, 
				 object,
				 signal_name,
				 GTK_SIGNAL_FUNC(on_nudge_bar_clicked),
				 nudge_info);
	return TRUE;
}

static gboolean gdam_turntable_nudge_beat_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	float amount;
	char* endp;
	NudgeInfo* nudge_info;
	if (handler_names[1] == NULL) {
		g_warning("%s needs a real-valued argument", handler_names[0]);
		return TRUE;
	}
	amount = strtod(handler_names[1], &endp);
	if (endp == handler_names[1] || *endp != '\0')
		return TRUE;

	nudge_info = g_new(NudgeInfo, 1);
	nudge_info->turntable = GDAM_TURNTABLE(skin->model);
	nudge_info->skin = GDAM_TURNTABLE_SKIN(skin);
	nudge_info->amount = amount;
	connect_button_with_mask(skin, 
				 object,
				 signal_name,
				 GTK_SIGNAL_FUNC(on_nudge_beat_clicked),
				 nudge_info);
	return TRUE;
}

static gboolean gdam_turntable_position_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkRange* range = GTK_RANGE(object);
	GtkAdjustment* adj = gtk_range_get_adjustment(range);
	GdamTurntable* turntable = GDAM_TURNTABLE(skin->model);
	gtk_signal_connect(GTK_OBJECT(adj),
			   "value_changed",
			   GTK_SIGNAL_FUNC(position_adj_changed),
			   turntable);
	gtk_signal_connect(object,
			   "button_press_event", 
			   GTK_SIGNAL_FUNC(posscroll_button_event),
			   turntable);
	gtk_signal_connect(object,
			   "button_release_event", 
			   GTK_SIGNAL_FUNC(posscroll_button_event),
			   turntable);
	gtk_signal_connect_object( GTK_OBJECT(turntable),
	      "set_position",
	      GTK_SIGNAL_FUNC(turntable_update_position_adj_value),
	      GTK_OBJECT(adj));
	gtk_signal_connect_object( GTK_OBJECT(turntable),
	      "update_length",
	      GTK_SIGNAL_FUNC(turntable_update_position_adj_upper),
	      GTK_OBJECT(adj));
	return TRUE;
}
static gboolean gdam_turntable_stop_button_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(
			skin, 
			object,
			signal_name,
			GTK_SIGNAL_FUNC(on_stop_button_clicked),
			GDAM_TURNTABLE(skin->model));
	return TRUE;
}
static gboolean gdam_turntable_time_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkLabel* label = GTK_LABEL(object);
	gtk_signal_connect_object(GTK_OBJECT(skin->model),
			  "set_position",
			  GTK_SIGNAL_FUNC(turntable_update_pos_label),
			  GTK_OBJECT(label));
	return TRUE;
}
static gboolean gdam_turntable_play_button_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(skin, object,
	                         "pressed",
			         GTK_SIGNAL_FUNC(on_play_button_clicked),
				 skin);
	return TRUE;
}
static gboolean gdam_turntable_zero_button_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(skin, object,
	                         signal_name,
			GTK_SIGNAL_FUNC(on_zero_button_clicked),
			skin->model);
	return TRUE;
}
static gboolean gdam_turntable_beatnum_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkLabel* label = GTK_LABEL(object);
	gtk_signal_connect_object(
			      GTK_OBJECT(skin->model),
			      "update_beatnum",
			      GTK_SIGNAL_FUNC(turntable_update_beatnum),
			      GTK_OBJECT(label));
	return TRUE;
}
static gboolean gdam_turntable_start_button_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect(object, signal_name,
	                   GTK_SIGNAL_FUNC(on_start_button_clicked),
			   skin);
	return TRUE;
}

static gboolean gdam_turntable_tempo_data_drag_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gdam_widget_string_source_set(GTK_WIDGET(object),
				      on_tempo_string_drag,
				      "gdam-tempo-text",
				      skin->model);
	gtk_signal_connect(object,
			   "button_press_event",
		      	   GTK_SIGNAL_FUNC(on_tempo_press_event),
			   skin);
	return TRUE;
}

static gboolean gdam_turntable_jump_scrollbar_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkRange* range = GTK_RANGE(object);
	GtkAdjustment* adj = gtk_range_get_adjustment(range);
	GdamTurntable* turntable = GDAM_TURNTABLE(skin->model);
	gtk_signal_connect(GTK_OBJECT(adj),
			   "value_changed",
			   GTK_SIGNAL_FUNC(on_jump_adj_changed),
			   turntable);
	gtk_signal_connect(object,
			   "button_release_event", 
			   GTK_SIGNAL_FUNC(on_jump_button_event),
			   turntable);
	return TRUE;
}
static gboolean gdam_turntable_resample_scrollbar_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adjustment;
	GdamTurntableSkin* turntable_skin = GDAM_TURNTABLE_SKIN(skin);
	adjustment = util_get_widget_adjustment(GTK_WIDGET(object));
	g_return_val_if_fail(adjustment != NULL, TRUE);
	turntable_skin->resample_adj = adjustment;
	gtk_signal_connect(GTK_OBJECT(adjustment),
			   "value_changed",
			   GTK_SIGNAL_FUNC(resample_adj_changed),
			   skin->model);
	gtk_signal_connect(object,
			   "button_press_event",
			   GTK_SIGNAL_FUNC(resample_adj_button_event),
			   skin);
	gtk_signal_connect(object,
			   "button_release_event",
			   GTK_SIGNAL_FUNC(resample_adj_button_event),
			   skin);
	gtk_signal_connect(
		      GTK_OBJECT(skin->model),
		      "set_rate",
		      GTK_SIGNAL_FUNC(turntable_update_rate_adj),
		      skin);
	gdam_delayed_adjustment_init(adjustment);
	return TRUE;
}
static gboolean gdam_turntable_halve_cbutton_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(
			skin, 
			object,
			signal_name,
			GTK_SIGNAL_FUNC(on_halve_check_button_clicked),
			GDAM_TURNTABLE(skin->model));
	return TRUE;
}
static gboolean gdam_turntable_rate_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(GTK_OBJECT(skin->model),
			          "set_rate",
			          GTK_SIGNAL_FUNC(on_turntable_set_rate),
			          (GtkObject*) GTK_LABEL(object));
	return TRUE;
}
static gboolean gdam_turntable_bpm_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkLabel* label = GTK_LABEL(object);
	gtk_signal_connect_object(
			      GTK_OBJECT(skin->model),
			      "set_beat_info",
			      GTK_SIGNAL_FUNC(turntable_update_tempo),
			      (GtkObject*)label);
	return TRUE;
}
static gboolean gdam_turntable_anchor_rbutton_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkRadioButton* radio = GTK_RADIO_BUTTON(object);
	GdamTurntable* turntable = GDAM_TURNTABLE(skin->model);
	gdam_infomanager_add_radio_button(radio);
	turntable->anchored = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(object));
	gtk_signal_connect(object,
		   "clicked",
		   GTK_SIGNAL_FUNC(on_anchor_rbutton_toggled),
		   turntable);
	gtk_signal_connect(object,
		   "toggled",
		   GTK_SIGNAL_FUNC(on_anchor_rbutton_toggled),
		   turntable);
	return TRUE;
}
static gboolean gdam_turntable_sync_button_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask( skin, object,
	                          signal_name,
		                  GTK_SIGNAL_FUNC(on_sync_button_clicked),
		                  skin->model);
	return TRUE;
}
static gboolean gdam_turntable_info_button_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(
		skin, 
		object,
		signal_name,
		GTK_SIGNAL_FUNC(on_info_button_clicked),
		skin->model);
	return TRUE;
}
static gboolean gdam_turntable_tempo_button_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(
		skin, 
		object,
		signal_name,
		GTK_SIGNAL_FUNC(on_tempo_button_clicked),
		skin->model);
	return TRUE;
}
static gboolean gdam_turntable_preview_button_pressed_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(
		skin, 
		object,
		signal_name,
		GTK_SIGNAL_FUNC(on_sync_preview_button_clicked),
		skin->model);
	return TRUE;
}
static gboolean gdam_turntable_preview_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkLabel* label = GTK_LABEL(object);
	GdamTurntable* turntable = GDAM_TURNTABLE(skin->model);
	gtk_signal_connect_object(
		      GTK_OBJECT(turntable),
		      "update_anchored",
		      GTK_SIGNAL_FUNC(update_anchored_label),
		      GTK_OBJECT(label));
	gtk_signal_connect_object(
		      GTK_OBJECT(turntable),
		      "update_sync_preview",
		      GTK_SIGNAL_FUNC(update_sync_preview_label),
		      GTK_OBJECT(label));
	return TRUE;
}
static gboolean gdam_turntable_jump_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkLabel* label = GTK_LABEL(object);
	gtk_signal_connect_object(
		      GTK_OBJECT(skin->model),
		      "update_jump_preview",
		      GTK_SIGNAL_FUNC(turntable_update_jump_preview),
		      GTK_OBJECT(label));
	return TRUE;
}
static gboolean gdam_turntable_producer_container_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamTurntableSkin* turntable_skin = GDAM_TURNTABLE_SKIN(skin);
	gdam_debug("howdy... matched prod_cont in connect_func");
	turntable_skin->producer_container = GTK_CONTAINER(object);
	if (turntable_skin->producer_skin) {
		GdamSkin* submodel_skin = turntable_skin->producer_skin;
		GtkWidget* widget = submodel_skin->main_widget;
		gdam_set_container(turntable_skin->producer_container, widget);
	}
	return TRUE;
}
static gboolean gdam_turntable_set_producer_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamTurntableSkin* turntable_skin = GDAM_TURNTABLE_SKIN(skin);
	if (!turntable_skin->producer_container)
		turntable_skin->producer_container = GTK_CONTAINER(object);
	if (handler_names[1] == NULL) {
		g_warning("no name for %s", handler_names[0]);
		return TRUE;
	}
	if (num_handler_names < 3) {
		g_warning("too few arguments to `%s' in turntable",
			handler_names[0]);
		return TRUE;
	}
	gdam_turntable_skin_set_producer(turntable_skin,
					 handler_names[1],
					 handler_names[2],
					 num_handler_names - 3,
					 handler_names + 3);
	return TRUE;
}

static gboolean gdam_turntable_build_new_model     (GdamSkin*      skin,
					            GdamChannel*   channel,
						    int            num_options,
				                    char**         options)
{
	GdamTurntable* turntable = gdam_turntable_new(channel);
	skin->model = GDAM_MODEL(turntable);

	if (num_options < 2) {
#if 0
	/* XXX TODO: this perhaps should be required... */
		g_warning("gdam_turntable_build_new_model: "
				"too few options, no producer");
		return FALSE;
		g_warning("gdam_turntable_build_new_model: "
						"delaying set producer");
#endif
		return TRUE;
	}

	gdam_turntable_skin_set_producer(GDAM_TURNTABLE_SKIN(skin),
					 options[0],
					 options[1],
					 num_options - 2,
					 options + 2);

	return TRUE;
}

static void gdam_turntable_skin_synchronize(GdamView* comp_skin)
{
	GdamTurntableSkin* turntable_skin = GDAM_TURNTABLE_SKIN(comp_skin);
	GdamModel* model = GDAM_SKIN(comp_skin)->model;
	GdamTurntable* turntable = GDAM_TURNTABLE(model);
	GdamModel* producer = GDAM_MODEL(turntable->producer);
	GdamSkin* skin;
	if (producer == NULL)
		return;
	skin = gdam_skin_find_from_model(producer);
	turntable_skin->producer_skin = skin;
	if (turntable_skin->producer_container != NULL
	 && skin->main_widget != NULL) {
		gdam_set_container(turntable_skin->producer_container,
		                   skin->main_widget);
		gtk_widget_show(skin->main_widget);
	}
		          
}

static GdamSkinGenericConnect connect_infos[] = {
	{ "gdam_turntable_frame_drag_data_received",
          gdam_turntable_frame_drag_data_received_handler	},
	{ "gdam_turntable_nudge_sec",
          gdam_turntable_nudge_sec_handler			},
	{ "gdam_turntable_nudge_bar",
          gdam_turntable_nudge_bar_handler			},
	{ "gdam_turntable_nudge_beat",
          gdam_turntable_nudge_beat_handler			},
	{ "gdam_turntable_position_adjustment",
          gdam_turntable_position_adjustment_handler		},
	{ "gdam_turntable_stop_button_pressed",
          gdam_turntable_stop_button_pressed_handler		},
	{ "gdam_turntable_time_label",
          gdam_turntable_time_label_handler			},
	{ "gdam_turntable_play_button_pressed",
          gdam_turntable_play_button_pressed_handler		},
	{ "gdam_turntable_zero_button_pressed",
          gdam_turntable_zero_button_pressed_handler		},
	{ "gdam_turntable_beatnum_label",
          gdam_turntable_beatnum_label_handler			},
	{ "gdam_turntable_start_button_pressed",
          gdam_turntable_start_button_pressed_handler		},
	{ "gdam_turntable_tempo_data_drag",
          gdam_turntable_tempo_data_drag_handler		},
	{ "gdam_turntable_jump_scrollbar",
          gdam_turntable_jump_scrollbar_handler			},
	{ "gdam_turntable_resample_scrollbar",
          gdam_turntable_resample_scrollbar_handler		},
	{ "gdam_turntable_halve_cbutton_pressed",
          gdam_turntable_halve_cbutton_pressed_handler		},
	{ "gdam_turntable_rate_label",
          gdam_turntable_rate_label_handler			},
	{ "gdam_turntable_bpm_label",
          gdam_turntable_bpm_label_handler			},
	{ "gdam_turntable_anchor_rbutton_pressed",
          gdam_turntable_anchor_rbutton_pressed_handler		},
	{ "gdam_turntable_sync_button_pressed",
          gdam_turntable_sync_button_pressed_handler		},
	{ "gdam_turntable_info_button_pressed",
          gdam_turntable_info_button_pressed_handler		},
	{ "gdam_turntable_tempo_button_pressed",
          gdam_turntable_tempo_button_pressed_handler		},
	{ "gdam_turntable_preview_button_pressed",
	  gdam_turntable_preview_button_pressed_handler		},
	{ "gdam_turntable_preview_label",
	  gdam_turntable_preview_label_handler			},
	{ "gdam_turntable_jump_label",
	  gdam_turntable_jump_label_handler			},
	{ "gdam_turntable_producer_container",
	  gdam_turntable_producer_container_handler		},
	{ "gdam_turntable_set_producer",
	  gdam_turntable_set_producer_handler			}	};

static void gdam_turntable_skin_destroy(GtkObject* object){
	GdamTurntableSkin* skin = GDAM_TURNTABLE_SKIN(object);
	if (skin->timer_id)
		gtk_timeout_remove(skin->timer_id);
}

static void gdam_turntable_skin_class_init(GdamTurntableSkinClass* klass)
{
	GdamSkinClass* skin_class = GDAM_SKIN_CLASS(klass);
	skin_class->build_new_model = gdam_turntable_build_new_model;
	GDAM_VIEW_CLASS(klass)->synchronize
		= gdam_turntable_skin_synchronize;
	GTK_OBJECT_CLASS(klass)->destroy = gdam_turntable_skin_destroy;
	gdam_skin_class_register_generic_handlers(skin_class,
			                          G_N_ELEMENTS(connect_infos),
			                          connect_infos);
}
